from .base import Base
from .default import Default

from .gzip import Gzip
from .lz4 import LZ4
from .zlib import Zlib

from .jpeg import Jpeg
from .png import Png


def from_name(codec_name: str) -> Base:
    if codec_name is None:
        return Default()
    compresslevel = None
    if ":" in codec_name:
        comp = codec_name.split(":")
        assert (
            len(comp) == 2
        ), f"{codec_name} is invalid compress format, should be [type] or [type:level]"
        codec_name = comp[0]
        compresslevel = int(comp[1])
    if codec_name == "default":
        return Default()
    elif codec_name == "gzip":
        return Gzip(compresslevel or 9)
    elif codec_name == "zlib":
        return Zlib(compresslevel or -1)
    elif codec_name == "lz4":
        return LZ4(compresslevel or 1)
    elif codec_name == "jpeg":
        return Jpeg()
    elif codec_name == "png":
        return Png()
    else:
        raise Exception("Unknown Codec")



from hub import config
from hub.log import logger
from hub.client.base import HubHttpClient
from hub.exceptions import HubException


class AuthClient(HubHttpClient):
    """
    Authentication through rest api
    """

    def __init__(self):
        super(AuthClient, self).__init__()

    def check_token(self, access_token):
        auth = "Bearer {}".format(access_token)
        response = self.request(
            "GET", config.CHECK_TOKEN_REST_SUFFIX, headers={"Authorization": auth}
        )

        try:
            response_dict = response.json()
            is_valid = response_dict["is_valid"]
        except Exception as e:
            logger.error("Exception occured while validating token: {}.".format(e))
            raise HubException(
                "Error while validating the token. \
                                  Please try logging in using username ans password."
            )

        return is_valid

    def get_access_token(self, username, password):
        response = self.request(
            "GET",
            config.GET_TOKEN_REST_SUFFIX,
            json={"username": username, "password": password},
        )

        try:
            token_dict = response.json()
            token = token_dict["token"]
        except Exception as e:
            logger.error("Exception occured while getting token: {}.".format(e))
            raise HubException(
                "Error while loggin in. \
                                  Please try logging in using access token."
            )
        return token

    def register(self, username, email, password):
        self.request(
            "POST",
            config.GET_REGISTER_SUFFIX,
            json={"username": username, "email": email, "password": password},
        )

        # try:
        #    token_dict = response.json()
        # except Exception as e:
        #    logger.error("Exception occured while registering token: {}.".format(e))
        #    raise HubException("Error while registering in. {e}")



from typing import Tuple

from hub.features.features import Tensor


class Image(Tensor):
    """`FeatureConnector` for images
    Output:
    `tf.Tensor` of type `tf.uint8` and shape `[height, width, num_channels]`
    for BMP, JPEG, and PNG images

    Example:
    ```python
    image_tensor = Image(shape=(None, None, 1),
                                encoding_format='png')
    ```
    """

    def __init__(
        self,
        shape: Tuple[int, ...] = (None, None, 3),
        dtype="uint8",
        encoding_format: str = "png",
        max_shape: Tuple[int, ...] = None,
        chunks=True,
    ):
        """Construct the connector.
        Args:
        shape: tuple of ints or None, the shape of decoded image:
            (height, width, channels) where height and width can be None.
            Defaults to (None, None, 3).
        dtype: `uint16` or `uint8` (default).
            `uint16` can be used only with png encoding_format
        encoding_format: 'jpeg' or 'png' (default). Format to serialize np.ndarray
            images on disk.
        channels: list or tuple of the names of channels: ("nir", "red", "green",...)
        Raises:
        ValueError: If the shape, dtype or encoding formats are invalid
        """
        self._set_dtype(dtype)
        super(Image, self).__init__(shape, dtype, max_shape=max_shape, chunks=chunks)
        self._set_encoding_format(encoding_format)

    def _set_encoding_format(self, encoding_format):
        """Set the encoding format."""
        if encoding_format not in ("png", "jpeg"):
            raise ValueError("Not supported encoding format")
        self.encoding_format = encoding_format

    def _set_dtype(self, dtype):
        """Set the dtype."""
        if dtype not in ("uint8", "uint16"):
            raise ValueError(f"Not supported dtype for {self.__class__.__name__}")
        self.dtype = dtype

    def get_attr_dict(self):
        """Return class attributes."""
        return self.__dict__



from typing import Tuple

import hub.features.flat_tensor as flat_tensor

FlatTensor = flat_tensor.FlatTensor


class FeatureConnector:
    def _flatten(self) -> Tuple[FlatTensor]:
        raise NotImplementedError()

    # @property
    # def shape(self):
    #     raise NotImplementedError()

    # @property
    # def dtype(self):
    #     raise NotImplementedError()



import fsspec
import os
import json
from hub.store.metastore import MetaStorage


def test_meta_storage():
    os.makedirs("./data/test/test_meta_storage/internal_tensor", exist_ok=True)
    fs: fsspec.AbstractFileSystem = fsspec.filesystem("file")
    meta_map = fs.get_mapper("./data/test/test_meta_storage")
    meta_map["meta.json"] = bytes(json.dumps(dict()), "utf-8")
    fs_map = fs.get_mapper("./data/test/test_meta_storage/internal_tensor")
    MetaStorage("internal_tensor", fs_map, meta_map)



import click
from hub.log import logger
from hub import config
from hub.client.auth import AuthClient
from hub.client.token_manager import TokenManager
from hub.client.hub_control import HubControlClient


@click.command()
# @click.option('--token', is_flag=True, default=False,
#        help='Enter authentication tocken from {}'.format(config.GET_TOKEN_REST_SUFFIX))
@click.option("--username", "-u", default=None, help="Your Snark AI username")
@click.option("--password", "-p", default=None, help="Your Snark AI password")
def login(username, password):
    """ Logs in to Snark AI"""
    token = ""
    if token:
        logger.info("Token login.")
        logger.degug("Getting the token...")
        token = click.prompt(
            "Please paste the authentication token from {}".format(
                config.GET_TOKEN_REST_SUFFIX, type=str, hide_input=True
            )
        )
        token = token.strip()
        AuthClient.check_token(token)
    else:
        logger.info(
            "Please log in using Snark AI credentials. You can register at https://app.activeloop.ai "
        )
        if not username:
            logger.debug("Prompting for username.")
            username = click.prompt("Username", type=str)
        username = username.strip()
        if not password:
            logger.debug("Prompting for password.")
            password = click.prompt("Password", type=str, hide_input=True)
        password = password.strip()
        token = AuthClient().get_access_token(username, password)
    TokenManager.set_token(token)
    HubControlClient().get_credentials()
    logger.info("Login Successful.")


@click.command()
def logout():
    """ Logs out of Activeloop AI"""
    TokenManager.purge_token()


@click.command()
@click.option("--username", "-u", default=None, help="Your Activeloop AI username")
@click.option("--email", "-e", default=None, help="Your email")
@click.option("--password", "-p", default=None, help="Your Activeloop AI password")
def register(username, email, password):
    """ Register at of Activeloop AI"""
    if not username:
        logger.debug("Prompting for username.")
        username = click.prompt("Username", type=str)
    username = username.strip()
    if not email:
        logger.debug("Prompting for email.")
        email = click.prompt("Email", type=str)
    email = email.strip()
    if not password:
        logger.debug("Prompting for password.")
        password = click.prompt("Password", type=str, hide_input=True)
    password = password.strip()
    AuthClient().register(username, email, password)
    token = AuthClient().get_access_token(username, password)
    TokenManager.set_token(token)



import os

curr_path = os.path.dirname(os.path.abspath(__file__))

TOKEN_FILE_PATH = os.path.expanduser("~/.activeloop/token")
STORE_CONFIG_PATH = os.path.expanduser("~/.activeloop/store")
CACHE_FILE_PATH = os.path.expanduser("~/.activeloop/tmp")
AWSCRED_PATH = os.path.expanduser("~/.aws/credentials")

HUB_REST_ENDPOINT = "https://app.activeloop.ai"
HUB_DEV_REST_ENDPOINT = "https://app.dev.activeloop.ai"
HUB_LOCAL_REST_ENDPOINT = "http://localhost:5000"

DEFAULT_TIMEOUT = 170

GET_TOKEN_REST_SUFFIX = "/api/user/token"
GET_CREDENTIALS_SUFFIX = "/api/credentials"
GET_REGISTER_SUFFIX = "/api/user/register"
GET_DATASET_SUFFIX = "/api/dataset/get"
GET_DATASET_PATH_SUFFIX = "/api/dataset/get/path"

CREATE_DATASET_SUFFIX = "/api/dataset/create"
UPDATE_STATE_SUFFIX = "/api/dataset/state"
DATASET_SUFFIX = "/api/dataset/"
DISTRIBUTED = True



import json
from collections.abc import MutableMapping
import posixpath


class MetaStorage(MutableMapping):
    @classmethod
    def to_str(cls, obj):
        if isinstance(obj, memoryview):
            obj = obj.tobytes()
        if isinstance(obj, bytes):
            obj = obj.decode("utf-8")
        return obj

    def __init__(self, path, fs_map: MutableMapping, meta_map: MutableMapping):
        self._fs_map = fs_map
        self._meta = meta_map
        self._path = path

    def __getitem__(self, k: str) -> bytes:
        filename = posixpath.split(k)[1]
        if filename.startswith("."):
            return bytes(
                json.dumps(
                    json.loads(self.to_str(self._meta["meta.json"]))[k][self._path]
                ),
                "utf-8",
            )
        else:
            return self._fs_map[k]

    def get(self, k: str) -> bytes:
        filename = posixpath.split(k)[1]
        if filename.startswith("."):
            meta_ = self._meta.get("meta.json")
            if not meta_:
                return None
            meta = json.loads(self.to_str(meta_))
            metak = meta.get(k)
            if not metak:
                return None
            item = metak.get(self._path)
            return bytes(json.dumps(item), "utf-8") if item else None
        else:
            return self._fs_map.get(k)

    def __setitem__(self, k: str, v: bytes):
        filename = posixpath.split(k)[1]
        if filename.startswith("."):
            meta = json.loads(self.to_str(self._meta["meta.json"]))
            meta[k] = meta.get(k) or {}
            meta[k][self._path] = json.loads(self.to_str(v))
            self._meta["meta.json"] = bytes(json.dumps(meta), "utf-8")
        else:
            self._fs_map[k] = v

    def __len__(self):
        return len(self._fs_map) + 1

    def __iter__(self):
        yield ".zarray"
        yield from self._fs_map

    def __delitem__(self, k: str):
        filename = posixpath.split(k)[1]
        if filename.startswith("."):
            meta = json.loads(self.to_str(self._meta["meta.json"]))
            meta[k] = meta.get(k) or dict()
            meta[k][self._path] = None
            self._meta["meta.json"] = bytes(json.dumps(meta), "utf-8")
        else:
            del self._fs_map[k]

    def commit(self):
        self._meta.commit()
        self._fs_map.commit()


from hub.api.tensorview import TensorView
from hub.api.dataset_utils import slice_extract_info, slice_split
import collections.abc as abc


class DatasetView:
    def __init__(self, dataset=None, num_samples=None, offset=None):
        assert dataset is not None
        assert num_samples is not None
        assert offset is not None

        self.dataset = dataset
        self.num_samples = num_samples
        self.offset = offset

    def __getitem__(self, slice_):
        """Gets a slice or slices from datasetview"""
        if not isinstance(slice_, abc.Iterable) or isinstance(slice_, str):
            slice_ = [slice_]
        slice_ = list(slice_)
        subpath, slice_list = slice_split(slice_)
        slice_list = [0] + slice_list if self.num_samples == 1 else slice_list
        if not subpath:
            if len(slice_list) > 1:
                raise ValueError(
                    "Can't slice a dataset with multiple slices without subpath"
                )
            num, ofs = slice_extract_info(slice_list[0], self.num_samples)
            return DatasetView(
                dataset=self.dataset, num_samples=num, offset=ofs + self.offset
            )
        elif not slice_list:
            slice_ = slice(self.offset, self.offset + self.num_samples)
            if subpath in self.dataset._tensors.keys():
                return TensorView(dataset=self.dataset, subpath=subpath, slice_=slice_)
            return self._get_dictionary(self.dataset, subpath, slice=slice_)
        else:
            num, ofs = slice_extract_info(slice_list[0], self.num_samples)
            slice_list[0] = (
                ofs + self.offset
                if num == 1
                else slice(ofs + self.offset, ofs + self.offset + num)
            )
            if subpath in self.dataset._tensors.keys():
                return TensorView(
                    dataset=self.dataset, subpath=subpath, slice_=slice_list
                )
            if len(slice_list) > 1:
                raise ValueError("You can't slice a dictionary of Tensors")
            return self._get_dictionary(subpath, slice_list[0])

    def __setitem__(self, slice_, value):
        """"Sets a slice or slices with a value"""
        if not isinstance(slice_, abc.Iterable) or isinstance(slice_, str):
            slice_ = [slice_]
        slice_ = list(slice_)
        subpath, slice_list = slice_split(slice_)
        slice_list = [0] + slice_list if self.num_samples == 1 else slice_list
        if not subpath:
            raise ValueError("Can't assign to dataset sliced without subpath")
        elif not slice_list:
            slice_ = (
                self.offset
                if self.num_samples == 1
                else slice(self.offset, self.offset + self.num_samples)
            )
            self.dataset._tensors[subpath][slice_] = value  # Add path check
        else:
            num, ofs = (
                slice_extract_info(slice_list[0], self.num_samples)
                if isinstance(slice_list[0], slice)
                else (1, slice_list[0])
            )
            slice_list[0] = (
                slice(ofs + self.offset, ofs + self.offset + num)
                if num > 1
                else ofs + self.offset
            )
            self.dataset._tensors[subpath][slice_list] = value

    def _get_dictionary(self, subpath, slice_=None):
        """"Gets dictionary from dataset given incomplete subpath"""
        tensor_dict = {}
        subpath = subpath if subpath.endswith("/") else subpath + "/"
        for key in self.dataset._tensors.keys():
            if key.startswith(subpath):
                suffix_key = key[len(subpath) :]
                split_key = suffix_key.split("/")
                cur = tensor_dict
                for i in range(len(split_key) - 1):
                    if split_key[i] not in cur.keys():
                        cur[split_key[i]] = {}
                    cur = cur[split_key[i]]
                slice_ = slice_ if slice_ else slice(0, self.dataset.shape[0])
                cur[split_key[-1]] = TensorView(
                    dataset=self.dataset, subpath=key, slice_=slice_
                )
        if len(tensor_dict) == 0:
            raise KeyError(f"Key {subpath} was not found in dataset")
        return tensor_dict



from hub.store.cache import Cache

from hub.client.hub_control import HubControlClient
import configparser
from typing import MutableMapping, Tuple

import fsspec
import gcsfs
import os


def _connect(tag):
    """Connects to the backend and receives credentials"""

    creds = HubControlClient().get_config()
    dataset = HubControlClient().get_dataset_path(tag)

    # If dataset is in DB then return the path
    # Otherwise construct the path from the tag
    if dataset and "path" in dataset:
        path = dataset["path"]
    else:
        sub_tags = tag.split("/")
        # Get repository path from the cred location
        path = "/".join(creds["bucket"].split("/")[:-1])
        path = f"{path}/{sub_tags[0]}/{sub_tags[-1]}"
    return path, creds


def get_fs_and_path(url: str, token=None) -> Tuple[fsspec.AbstractFileSystem, str]:
    if url.startswith("s3://"):
        token = token or dict()
        token = read_aws_creds(token) if isinstance(token, str) else token
        return (
            fsspec.filesystem(
                "s3",
                key=token.get("aws_access_key_id"),
                secret=token.get("aws_secret_access_key"),
                token=token.get("aws_session_token"),
            ),
            url[5:],
        )
    elif url.startswith("gcs://"):
        return gcsfs.GCSFileSystem(token=token), url[6:]
    elif url.startswith("abs://"):
        # TODO: Azure
        raise NotImplementedError()
    elif (
        url.startswith("../")
        or url.startswith("./")
        or url.startswith("/")
        or url.startswith("~/")
    ):
        return fsspec.filesystem("file"), url
    else:
        # TOOD check if url is username/dataset:version
        url, creds = _connect(url)
        fs = fsspec.filesystem(
            "s3",
            key=creds["access_key"],
            secret=creds["secret_key"],
            token=creds["session_token"],
            client_kwargs={
                "endpoint_url": creds["endpoint"],
                "region_name": creds["region"],
            },
        )

        return (fs, url)


def read_aws_creds(filepath: str):
    parser = configparser.ConfigParser()
    parser.read(filepath)
    return {section: dict(parser.items(section)) for section in parser.sections()}


def _get_storage_map(fs, path):
    return StorageMapWrapperWithCommit(fs.get_mapper(path, check=False, create=False))


def get_storage_map(fs, path, memcache=2 ** 26, lock=True):
    store = _get_storage_map(fs, path)
    cache_path = os.path.join("~/.activeloop/cache/", path)
    return Cache(store, memcache, path=cache_path, lock=lock)


class StorageMapWrapperWithCommit(MutableMapping):
    def __init__(self, map):
        self._map = map
        self.root = self._map.root

    def __getitem__(self, slice_):
        return self._map[slice_]

    def __setitem__(self, slice_, value):
        self._map[slice_] = value

    def __delitem__(self, slice_):
        del self._map[slice_]

    def __len__(self):
        return len(self._map)

    def __iter__(self):
        yield from self._map

    def commit(self):
        pass



# from typing import Tuple

from hub.features.features import Tensor


class BBox(Tensor):
    """`FeatureConnector` for a normalized bounding box.
    Output:
    bbox: Tensor of type `float32` and shape `[4,]` which contains the
          normalized coordinates of the bounding box `[ymin, xmin, ymax, xmax]`
    """
    def __init__(self, dtype=None, chunks=True):
        """Construct the connector.
        Args:
        dtype: dtype of bbox coordinates.
        Default: 'float32'
        """
        if not dtype:
            dtype = 'float32'
        super(BBox, self).__init__(shape=(4,), dtype=dtype, chunks=chunks)

    def get_attr_dict(self):
        """Return class attributes."""
        return self.__dict__



import json
import os
import time
from hub import config
from hub.client.base import HubHttpClient
from hub.client.auth import AuthClient
from pathlib import Path

from hub.exceptions import NotFoundException


class HubControlClient(HubHttpClient):
    """
    Controlling Hub through rest api
    """

    def __init__(self):
        super(HubControlClient, self).__init__()
        self.details = self.get_config()

    def get_dataset_path(self, tag):
        try:
            dataset = self.request(
                "GET",
                config.GET_DATASET_PATH_SUFFIX,
                params={"tag": tag},
                endpoint=config.HUB_REST_ENDPOINT,
            ).json()
        except NotFoundException:
            dataset = None

        return dataset

    def get_credentials(self):

        if self.auth_header is None:
            token = AuthClient().get_access_token(username="public", password="")
            self.auth_header = f"Bearer {token}"

        r = self.request(
            "GET", config.GET_CREDENTIALS_SUFFIX, endpoint=config.HUB_REST_ENDPOINT,
        ).json()

        details = {
            "_id": r["_id"],
            "region": r["region"],
            "session_token": r["session_token"],
            "access_key": r["access_key"],
            "secret_key": r["secret_key"],
            "endpoint": r["endpoint"],
            "expiration": r["expiration"],
            "bucket": r["bucket"],
        }

        self.save_config(details)
        return details

    def get_config(self, reset=False):

        if not os.path.isfile(config.STORE_CONFIG_PATH) or self.auth_header is None:
            self.get_credentials()

        with open(config.STORE_CONFIG_PATH, "r") as file:
            details = file.readlines()
            details = json.loads("".join(details))

        if float(details["expiration"]) < time.time() - 36000 or reset:
            details = self.get_credentials()
        return details

    def save_config(self, details):
        path = Path(config.STORE_CONFIG_PATH)
        os.makedirs(path.parent, exist_ok=True)
        with open(config.STORE_CONFIG_PATH, "w") as file:
            file.writelines(json.dumps(details))

    def create_dataset_entry(self, username, dataset_name, meta, public=True):
        try:
            tag = f"{username}/{dataset_name}"
            repo = f"public/{username}" if public else f"private/{username}"
            print(tag, repo, public)
            self.request(
                "POST",
                config.CREATE_DATASET_SUFFIX,
                json={
                    "tag": tag,
                    "repository": repo,
                    "public": public,
                    "rewrite": True
                },
                endpoint=config.HUB_REST_ENDPOINT,
            ).json()
        except Exception as e:
            print("Unable to create Dataset entry")
            print(e)

    def update_dataset_state(self, username, dataset_name, state, progress=0):
        try:
            tag = f"{username}/{dataset_name}"
            self.request(
                "POST",
                config.UPDATE_STATE_SUFFIX,
                json={
                    "tag": tag,
                    "state": state,
                    "progress": progress,
                },
                endpoint=config.HUB_REST_ENDPOINT,
            ).json()
        except Exception as e:
            print("Unable to update Dataset entry state")
            print(e)

    def delete_dataset_entry(self, username, dataset_name):
        try:
            tag = f"{username}/{dataset_name}"
            suffix = f"{config.DATASET_SUFFIX}/{tag}"
            self.request(
                "DELETE",
                suffix,
                endpoint=config.HUB_REST_ENDPOINT,
            ).json()
        except Exception as e:
            print("Unable to delete Dataset entry")
            print(e)



from typing import Tuple

from hub.features.image import Image
from hub.features.sequence import Sequence


class Video(Sequence):
    """`FeatureConnector` for videos, encoding frames individually on disk.

    The connector accepts as input a 4 dimensional `uint8` array
    representing a video.
    Output:
        video: Tensor of type `uint8` and shape
        [num_frames, height, width, channels], where channels must be 1 or 3
    """
    def __init__(
        self,
        shape: Tuple[int, ...] = None,
        dtype: str = 'uint8',
        encoding_format: str = 'png',
        chunks=True
    ):
        """Initializes the connector.
            Args:
            shape: tuple of ints, the shape of the video (num_frames, height, width,
                channels), where channels is 1 or 3.
            encoding_format: The video is stored as a sequence of encoded images.
                You can use any encoding format supported by Image.
            dtype: `uint16` or `uint8` (default).
            Raises:
            ValueError: If the shape, dtype or encoding formats are invalid
        """
        super(Video, self).__init__(feature=Image(shape=shape[1:], dtype=dtype,
                                    encoding_format=encoding_format), length=shape[0], chunks=chunks)

    def get_attr_dict(self):
        """Return class attributes."""
        return self.__dict__



from hub.store.cache import Cache
import zarr
import time
import posixpath


class SlowStore(zarr.MemoryStore):
    def __init__(self, **kwargs):
        super(SlowStore, self).__init__(**kwargs)

    def __getitem__(self, key, **kwargs):
        time.sleep(0.001)
        return super(SlowStore, self).__getitem__(key, **kwargs)

    def __setitem__(self, key, value, **kwargs):
        super(SlowStore, self).__setitem__(key, value, **kwargs)


def test_cache():
    store = SlowStore()
    store = Cache(store, max_size=1000000)

    for i in range(10):
        z = zarr.zeros(
            (1000, 1000),
            chunks=(100, 100),
            path=posixpath.realpath(f"./data/test/test_cache/first{i}"),
            store=store,
            overwrite=True,
        )

        z[...] = i
        store.invalidate()

        t1 = time.time()
        z[...]
        t2 = time.time()
        z[...]
        t3 = time.time()
        assert z[0, 0] == i
        # print(t2 - t1, t3 - t2)
        assert t2 - t1 > t3 - t2
    store.commit()


if __name__ == "__main__":
    test_cache()


import collections.abc as abc
from hub.api.dataset_utils import slice_split


class TensorView:
    def __init__(
        self,
        dataset=None,
        subpath=None,
        slice_=None,
    ):

        assert dataset is not None
        assert subpath is not None

        self.dataset = dataset
        self.subpath = subpath

        if isinstance(slice_, int) or isinstance(slice_, slice):
            self.slice_ = [slice_]
        elif isinstance(slice_, tuple) or isinstance(slice_, list):
            self.slice_ = list(slice_)
        self.nums = []
        self.offsets = []
        for it in self.slice_:
            if isinstance(it, int):
                self.nums.append(1)
                self.offsets.append(it)
            elif isinstance(it, slice):
                ofs = it.start if it.start else 0
                num = it.stop - ofs if it.stop else None
                self.nums.append(num)
                self.offsets.append(ofs)
        self.dtype = self.dtype_from_path(subpath)
        tensor_shape = self.dtype.shape if hasattr(self.dtype, "shape") else (1,)
        self.shape = self.make_shape(tensor_shape)

    def numpy(self):
        """Gets the value from tensorview"""
        return self.dataset._tensors[self.subpath][self.slice_]

    def compute(self):
        """Gets the value from tensorview"""
        return self.numpy()

    def __getitem__(self, slice_):
        """Gets a slice or slices from tensorview"""
        if not isinstance(slice_, abc.Iterable) or isinstance(slice_, str):
            slice_ = [slice_]
        slice_ = list(slice_)
        slice_ = self.slice_fill(slice_)
        subpath, slice_list = slice_split(slice_)
        if subpath:
            raise ValueError("Can't slice a Tensor with string")
        else:
            new_nums = self.nums.copy()
            new_offsets = self.offsets.copy()
            if len(new_nums) < len(slice_list):
                new_nums.extend([None] * (len(slice_list) - len(new_nums)))
                new_offsets.extend([0] * (len(slice_list) - len(new_offsets)))
            for i in range(len(slice_list)):
                slice_list[i] = self._combine(
                    slice_list[i], new_nums[i], new_offsets[i]
                )
            for i in range(len(slice_list), len(new_nums)):
                cur_slice = slice(new_offsets[i], new_offsets[i] + new_nums[i]) if new_nums[i] > 1 else new_offsets[i]
                slice_list.append(cur_slice)
            return TensorView(dataset=self.dataset, subpath=self.subpath, slice_=slice_list)

    def __setitem__(self, slice_, value):
        """"Sets a slice or slices with a value"""
        if not isinstance(slice_, abc.Iterable) or isinstance(slice_, str):
            slice_ = [slice_]
        slice_ = list(slice_)
        slice_ = self.slice_fill(slice_)
        subpath, slice_list = slice_split(slice_)
        if subpath:
            raise ValueError("Can't slice a Tensor with multiple slices without subpath")
        else:
            new_nums = self.nums.copy()
            new_offsets = self.offsets.copy()
            if len(new_nums) < len(slice_list):
                new_nums.extend([None] * (len(slice_list) - len(new_nums)))
                new_offsets.extend([0] * (len(slice_list) - len(new_offsets)))
            for i in range(len(slice_list)):
                slice_list[i] = self._combine(slice_[i], new_nums[i], new_offsets[i])
            for i in range(len(slice_list), len(new_nums)):
                cur_slice = slice(new_offsets[i], new_offsets[i] + new_nums[i]) if new_nums[i] > 1 else new_offsets[i]
                slice_list.append(cur_slice)
            self.dataset._tensors[self.subpath][slice_list] = value

    def _combine(self, slice_, num=None, ofs=0):
        "combines a slice_ with the current num and offset present in tensorview"
        if isinstance(slice_, int):
            self.check_slice_bounds(num=num, start=slice_)
            return ofs + slice_
        elif isinstance(slice_, slice):
            self.check_slice_bounds(
                num=num, start=slice_.start, stop=slice_.stop, step=slice_.step
            )
            if slice_.start is None and slice_.stop is None:
                return slice(ofs, None) if num is None else slice(ofs, ofs + num)
            elif slice_.start is not None and slice_.stop is None:
                return (
                    slice(ofs + slice_.start, None)
                    if num is None
                    else slice(ofs + slice_.start, ofs + num)
                )
            elif slice_.start is None and slice_.stop is not None:
                return slice(ofs, ofs + slice_.stop)
            else:
                return slice(ofs + slice_.start, ofs + slice_.stop)
        else:
            raise TypeError(
                "type {} isn't supported in dataset slicing".format(type(slice_))
            )

    def check_slice_bounds(self, num=None, start=None, stop=None, step=None):
        "checks whether the bounds of slice are in limits"
        if step and step < 0:  # negative step not supported
            raise ValueError("Negative step not supported in dataset slicing")
        if num and ((start and start >= num) or (stop and stop > num)):
            raise IndexError(
                "index out of bounds for dimension with length {}".format(num)
            )
        if start and stop and start > stop:
            raise IndexError("start index is greater than stop index")

    def dtype_from_path(self, path):
        "gets the dtype of the Tensorview by traversing the schema"
        path = path.split('/')
        cur_type = self.dataset.schema.dict_
        for subpath in path[1:-1]:
            cur_type = cur_type[subpath]
            cur_type = cur_type.dict_
        return cur_type[path[-1]]

    def slice_fill(self, slice_):
        "fills the slice with zeroes for the dimensions that have single elements"
        new_slice_ = []
        offset = 0
        for num in self.nums:
            if num == 1:
                new_slice_.append(0)
            else:
                new_slice_.append(slice_[offset])
                offset += 1
        new_slice_ = new_slice_ + slice_[offset:]
        return new_slice_

    def make_shape(self, shape):
        "combines the Tensorview slice and underlying shape to get the shape represented by it"
        shape = []
        shape.append(self.nums[0])
        for i in range(len(shape)):
            if i + 1 < len(self.nums):
                shape.append(self.nums[i + 1])
            else:
                shape.append(shape[i])
        final_shape = [dim for dim in shape if dim != 1]
        return tuple(final_shape)



import typing


class Tensor:
    def __getitem__(self, slice_) -> typing.Union["Tensor", typing.Any]:
        pass

    def __setitem__(self, slice_, value):
        pass

    def __repr__(self) -> str:
        pass

    def __str__(self) -> str:
        pass



import numpy


class Base:
    def encode(self, array: numpy.ndarray) -> bytes:
        raise NotImplementedError()

    def decode(self, bytes: bytes) -> numpy.ndarray:
        raise NotImplementedError()



from typing import Tuple


class FlatTensor:
    def __init__(
        self, path: str, shape: Tuple[int, ...], max_shape: Tuple[int, ...], dtype: str
    ):
        self.path = path
        self.shape = shape
        self.max_shape = max_shape
        self.dtype = dtype



from typing import Tuple

from hub.features.features import Tensor


class Polygon(Tensor):
    """`FeatureConnector` for polygon
    """
    def __init__(
        self,
        shape: Tuple[int, ...] = None,
        max_shape: Tuple[int, ...] = None,
        chunks=True
    ):
        """Constructs a Polygon FeatureConnector.
        Args:
        shape: tuple of ints or None, i.e (None, 2)

        Example:
        ```python
        polygon_tensor = Polygon(shape=(10, 2))
        polygon_tensor = Polygon(shape=(None, 2))
        ```

        Raises:
        ValueError: If the shape is invalid
        """
        self._check_shape(shape)
        super(Polygon, self).__init__(shape, dtype='uint32', max_shape=max_shape, chunks=chunks)

    def _check_shape(self, shape):
        if len(shape) != 2 or shape[-1] != 2:
            raise ValueError("Wrong polygon shape provided")

    def get_attr_dict(self):
        """Return class attributes."""
        return self.__dict__



import collections.abc as abc
import json
import math

import numpy as np
import zarr
import numcodecs

from hub.store.nested_store import NestedStore

from hub.exceptions import DynamicTensorNotFoundException, ValueShapeError
from hub.api.dataset_utils import slice_extract_info


def _tuple_product(tuple_):
    res = 1
    for t in tuple_:
        res *= t
    return res


def _determine_chunksizes(shape, dtype, block_size=2 ** 24):
    """
    Autochunking of tensors
    Chunk is determined by 16MB blocks keeping left dimensions inside a chunk
    Dimensions from left are kept until 16MB block is filled

    Parameters
    ----------
    shape: tuple
        the shape of the whole array
    dtype: type
        the type of the element (int, float)
    block_size: int (optional)
        how big the chunk size should be. Default to 16MB
    """

    sz = np.dtype(dtype).itemsize
    elem_count_in_chunk = block_size / sz

    # Get left most part which will be left static inside the chunk
    a = list(shape)
    a.reverse()
    left_part = shape
    prod = 1
    for i, dim in enumerate(a):
        prod *= dim
        if elem_count_in_chunk < prod:
            left_part = shape[-i:]
            break

    # If the tensor is smaller then the chunk size return
    if len(left_part) == len(shape):
        return list(left_part)

    # Get the middle chunk size of dimension
    els = math.ceil(elem_count_in_chunk / _tuple_product(left_part))

    # Contruct the chunksize shape
    chunksize = [els] + list(left_part)
    if len(chunksize) < len(shape):
        chunksize = [1] * (len(shape) - len(chunksize)) + chunksize
    return list(chunksize)


class DynamicTensor:
    """Class for handling dynamic tensor

    This class adds dynamic nature to storage tensor.
    The shape of tensor depends only on the index of the first dim.
    """

    # TODO Make first dim is extensible as well
    def __init__(
        self,
        fs_map: str,
        mode: str = "r",
        shape=None,
        max_shape=None,
        dtype="float64",
        chunks=None,
    ):
        """Constructor
        Parameters
        ----------
        fs_map : MutableMap
            Maps filesystem to MutableMap
        mode : str
            Mode in which tensor is opened (default is "r"), can be used to overwrite or append
        shape : Tuple[int | None]
            Shape of tensor, (must be specified) can contains Nones meaning the shape might change
        max_shape: Tuple[int | None]
            Maximum possible shape of the tensor (must be specified)
        dtype : str
            Numpy analog dtype for this tensor
        chunks : Tuple[int] | True
            How to split the tensor into chunks (files) (default is True)
            If chunks=True then chunksize will automatically be detected

        """
        exist_ = fs_map.get(".hub.dynamic_tensor")
        # if not exist_ and len(fs_map) > 0 and "w" in mode:
        #     raise OverwriteIsNotSafeException()
        exist = False if "w" in mode else exist_ is not None
        if "r" in mode and not exist:
            raise DynamicTensorNotFoundException()

        synchronizer = None
        # syncrhonizer = zarr.ProcessSynchronizer("~/activeloop/sync/example.sync")
        # if tensor exists and mode is read or append
        if ("r" in mode or "a" in mode) and exist:
            meta = json.loads(fs_map.get(".hub.dynamic_tensor").decode("utf-8"))
            shape = meta["shape"]
            self._dynamic_dims = get_dynamic_dims(shape)
            self._storage_tensor = zarr.open_array(
                store=fs_map, mode=mode, synchronizer=synchronizer
            )
            self._dynamic_tensor = (
                zarr.open_array(
                    NestedStore(fs_map, "--dynamic--"),
                    mode=mode,
                    synchronizer=synchronizer,
                )
                if self._dynamic_dims
                else None
            )
        # else we need to create or overwrite the tensor
        else:
            self._dynamic_dims = get_dynamic_dims(shape)
            self._storage_tensor = zarr.zeros(
                max_shape,
                dtype=dtype,
                chunks=chunks or _determine_chunksizes(max_shape, dtype),
                store=fs_map,
                overwrite=("w" in mode),
                object_codec=numcodecs.Pickle() if str(dtype) == "object" else None,
                synchronizer=synchronizer,
            )
            self._dynamic_tensor = (
                zarr.zeros(
                    shape=(max_shape[0], len(self._dynamic_dims)),
                    mode=mode,
                    dtype=np.int32,
                    store=NestedStore(fs_map, "--dynamic--"),
                    synchronizer=synchronizer,
                )
                if self._dynamic_dims
                else None
            )
            fs_map[".hub.dynamic_tensor"] = bytes(json.dumps({"shape": shape}), "utf-8")
        self.shape = shape
        self.max_shape = self._storage_tensor.shape
        self.dtype = self._storage_tensor.dtype
        assert len(self.shape) == len(self.max_shape)
        for item in self.max_shape:
            assert item is not None
        for item in zip(self.shape, self.max_shape):
            if item[0] is not None:
                # FIXME throw an error and explain whats wrong
                assert item[0] == item[1]

    def __getitem__(self, slice_):
        """Gets a slice or slices from tensor"""
        if not isinstance(slice_, abc.Iterable):
            slice_ = [slice_]
        slice_ = list(slice_)
        # real_shapes is dynamic shapes based on first dim index, only dynamic dims are stored, static ones are ommitted
        if self._dynamic_tensor:
            real_shapes = self._dynamic_tensor[slice_[0]]
        else:
            real_shapes = None
        # Extend slice_ to dim count
        slice_ += [slice(0, None, 1) for i in self.max_shape[len(slice_) :]]
        slice_ = self._get_slice(slice_, real_shapes)
        return self._storage_tensor[slice_]

    def __setitem__(self, slice_, value):
        """Sets a slice or slices with a value"""
        if not isinstance(slice_, abc.Iterable):
            slice_ = [slice_]
        slice_ = list(slice_)
        if self._dynamic_tensor:
            self.set_shape(slice_, value)
        slice_ += [slice(0, None, 1) for i in self.max_shape[len(slice_) :]]
        real_shapes = self._dynamic_tensor[slice_[0]] if self._dynamic_tensor else None
        slice_ = self._get_slice(slice_, real_shapes)

        if None not in self.shape and not all([isinstance(sh, int) for sh in slice_]):
            expected_value_shape = tuple([len(range(*slice_shape.indices(self.shape[i])))                                       
                                        for i, slice_shape in enumerate(slice_) 
                                        if not isinstance(slice_shape, int)])
            if type(value) in (np.ndarray, list) and np.array(value).shape != expected_value_shape:
                raise ValueShapeError(expected_value_shape, value.shape)
        self._storage_tensor[slice_] = value

    def get_shape(self, slice_):
        """Gets the shape of the slice from tensor"""
        if self._dynamic_tensor is None:
            return self.shape
        if isinstance(slice_[0], slice):
            num, ofs = slice_extract_info(slice_[0], self.shape[0])
            slice_[0] = ofs if num == 1 else slice_[0]
        if isinstance(slice_[0], int):
            final_shape = []
            shape_offset = 0
            for i in range(1, len(self.shape)):
                if i < len(slice_):
                    if isinstance(slice_[i], slice):
                        final_shape.append(slice_[i].stop - slice_[i].start)
                    shape_offset += 1
                elif self.shape[i] is not None:
                    final_shape.append(self.shape[i])
                elif shape_offset < len(self._dynamic_tensor[slice_[0]]):
                    final_shape.append(self._dynamic_tensor[slice_[0]][shape_offset])
                    shape_offset += 1
            return tuple(final_shape)
        else:
            raise ValueError("Getting shape across multiple dimensions isn't supported")

    def set_shape(self, slice_, value):
        """Sets the shape of the slice of tensor"""
        if isinstance(slice_[0], slice):
            num, ofs = slice_extract_info(slice_[0], self.shape[0])
            slice_[0] = ofs if num == 1 else slice_[0]
        if isinstance(slice_[0], int):
            value_shape = list(value.shape) if isinstance(value, np.ndarray) else [1]
            new_shape = []
            shape_offset = 0
            for i in range(1, len(self.shape)):
                if self.shape[i] is None:
                    if i < len(slice_):
                        if isinstance(slice_[i], slice):
                            sl = slice_[i].stop
                            shape_offset += 1
                        else:
                            sl = slice_[i] + 1
                        new_shape.append(sl)
                    else:
                        new_shape.append(value_shape[shape_offset])
                        shape_offset += 1
                elif i < len(slice_) and isinstance(slice_[i], slice):
                    shape_offset += 1
            self._dynamic_tensor[slice_[0]] = np.maximum(
                self._dynamic_tensor[slice_[0]], new_shape
            )
        else:
            raise ValueError("Setting shape across multiple dimensions isn't supported")

    def _get_slice(self, slice_, real_shapes):
        # Makes slice_ which is uses relative indices (ex [:-5]) into precise slice_ (ex [10:40])
        slice_ = list(slice_)
        if real_shapes is not None:
            for r, i in enumerate(self._dynamic_dims):
                if isinstance(slice_[i], int) and slice_[i] < 0:
                    slice_[i] += real_shapes[r]
                elif isinstance(slice_[i], slice) and (
                    slice_[i].stop is None or slice_[i].stop < 0
                ):
                    slice_[i] = slice_stop_changed(
                        slice_[i], (slice_[i].stop or 0) + real_shapes[r]
                    )
        return tuple(slice_)

    @classmethod
    def _get_slice_upper_boundary(cls, slice_):
        if isinstance(slice_, slice):
            return slice_.stop
        else:
            assert isinstance(slice_, int)
            return slice_ + 1

    @property
    def chunksize(self):
        """
        Get chunk shape of the array
        """
        return self._storage_tensor.chunks

    def _get_chunking_dim(self):
        for i, d in enumerate(self.chunksize):
            if d != 1:
                return i, self.shape[i], self.chunksize[i]
        return 0, self.shape[0], self.chunksize[0]

    def chunk_slice_iterator(self):
        """
        Get an iterator over chunk coordinates
        """
        # FIXME assume chunking is done in one dimension
        nth, shpd, chnkd = self._get_chunking_dim()
        n_els = int(shpd / chnkd)
        for i in range(n_els):
            yield [1] * nth + [slice(i * chnkd, (i + 1) * chnkd)]

    def chunk_iterator(self):
        """
        Get an iterator over chunks
        """
        slices = self.chunk_slice_iterator()
        for slice_chunk in slices:
            yield self.__getitem__(*slice_chunk)

    def commit(self):
        self._storage_tensor.store.commit()
        if self._dynamic_tensor:
            self._dynamic_tensor.store.commit()


def get_dynamic_dims(shape):
    return [i for i, s in enumerate(shape) if s is None]


def slice_stop_changed(slice_, new_stop):
    return slice(slice_.start, new_stop, slice_.step)



def slice_split(slice_):
    """Splits a slice into subpath and list of slices"""
    path = ""
    list_slice = []
    for sl in slice_:
        if isinstance(sl, str):
            path += sl if sl.startswith("/") else "/" + sl
        elif isinstance(sl, int) or isinstance(sl, slice):
            list_slice.append(sl)
        else:
            raise TypeError(
                "type {} isn't supported in dataset slicing".format(type(sl))
            )
    return path, list_slice


def slice_extract_info(slice_, num):
    """Extracts number of samples and offset from slice"""
    if isinstance(slice_, int):
        slice_ = slice_ + num if slice_ < 0 else slice_
        if slice_ >= num or slice_ < 0:
            raise IndexError(
                "index out of bounds for dimension with length {}".format(num)
            )
        return (1, slice_)

    if slice_.step is not None and slice_.step < 0:  # negative step not supported
        raise ValueError("Negative step not supported in dataset slicing")
    offset = 0
    if slice_.start is not None:
        slice_ = (
            slice(slice_.start + num, slice_.stop) if slice_.start < 0 else slice_
        )  # make indices positive if possible
        if slice_.start < 0 or slice_.start >= num:
            raise IndexError(
                "index out of bounds for dimension with length {}".format(num)
            )
        offset = slice_.start
    if slice_.stop is not None:
        slice_ = (
            slice(slice_.start, slice_.stop + num) if slice_.stop < 0 else slice_
        )  # make indices positive if possible
        if slice_.stop < 0 or slice_.stop > num:
            raise IndexError(
                "index out of bounds for dimension with length {}".format(num)
            )
    if slice_.start is not None and slice_.stop is not None:
        num = 0 if slice_.stop < slice_.start else slice_.stop - slice_.start
    elif slice_.start is None and slice_.stop is not None:
        num = slice_.stop
    elif slice_.start is not None and slice_.stop is None:
        num = num - slice_.start
    return num, offset



import logging
import sys

logger = logging.getLogger('hub')


def configure_logger(debug=0):
    log_level = logging.DEBUG if debug == 1 else logging.INFO
    logging.basicConfig(format='%(message)s',
                        level=log_level,
                        stream=sys.stdout)


configure_logger(0)



from click import ClickException


class OutOfBoundsError(Exception):
    """Raised upon finding a missing chunk."""

    pass


class AlignmentError(Exception):
    """Raised when there is an Alignment error."""

    pass


class IncompatibleShapes(Exception):
    """Shapes do not match"""

    pass


class IncompatibleBroadcasting(Exception):
    """Broadcasting issue"""

    pass


class IncompatibleTypes(Exception):
    """Types can not cast"""

    pass


class WrongTypeError(Exception):
    """Types is not supported"""

    pass


class NotAuthorized(Exception):
    """Types is not supported"""

    pass


class NotFound(Exception):
    """When Info could not be found for array"""

    pass


class FileSystemException(Exception):
    """Error working with local file system"""

    pass


class S3Exception(Exception):
    """Error working with AWS"""

    pass


class S3CredsParseException(Exception):
    """Can't parse AWS creds"""

    pass


class HubException(ClickException):
    def __init__(self, message=None, code=None):
        super(HubException, self).__init__(message)


class AuthenticationException(HubException):
    def __init__(self, message="Authentication failed. Please login again."):
        super(AuthenticationException, self).__init__(message=message)


class AuthorizationException(HubException):
    def __init__(self, response):
        try:
            message = response.json()["message"]
        except (KeyError, AttributeError):
            message = "You are not authorized to access this resource on Snark AI."
        super(AuthorizationException, self).__init__(message=message)


class NotFoundException(HubException):
    def __init__(
        self,
        message="The resource you are looking for was not found. Check if the name or id is correct.",
    ):
        super(NotFoundException, self).__init__(message=message)


class BadRequestException(HubException):
    def __init__(self, response):
        try:
            message = "One or more request parameters is incorrect\n%s" % str(
                response.json()["message"]
            )
        except (KeyError, AttributeError):
            message = "One or more request parameters is incorrect, %s" % str(
                response.content
            )
        super(BadRequestException, self).__init__(message=message)


class OverLimitException(HubException):
    def __init__(
        self,
        message="You are over the allowed limits for this operation. Consider upgrading your account.",
    ):
        super(OverLimitException, self).__init__(message=message)


class ServerException(HubException):
    def __init__(self, message="Internal Snark AI server error."):
        super(ServerException, self).__init__(message=message)


class BadGatewayException(HubException):
    def __init__(self, message="Invalid response from Snark AI server."):
        super(BadGatewayException, self).__init__(message=message)


class GatewayTimeoutException(HubException):
    def __init__(self, message="Snark AI server took too long to respond."):
        super(GatewayTimeoutException, self).__init__(message=message)


class WaitTimeoutException(HubException):
    def __init__(self, message="Timeout waiting for server state update."):
        super(WaitTimeoutException, self).__init__(message=message)


class LockedException(HubException):
    def __init__(self, message="Resource locked."):
        super(LockedException, self).__init__(message=message)


class HubDatasetNotFoundException(HubException):
    def __init__(self, response):
        message = f"The dataset with tag {response} was not found"
        super(HubDatasetNotFoundException, self).__init__(message=message)


class PermissionException(HubException):
    def __init__(self, response):
        message = f"No permision to store the dataset at {response}"
        super(PermissionException, self).__init__(message=message)

class ShapeArgumentNotFoundException(HubException):
    def __init__(self):
        message = f"Parameter 'shape' should be provided for Dataset creation."
        super(HubException, self).__init__(message=message)

class SchemaArgumentNotFoundException(HubException):
    def __init__(self):
        message = f"Parameter 'schema' should be provided for Dataset creation."
        super(HubException, self).__init__(message=message)

class ValueShapeError(HubException):
    def __init__(self, correct_shape, wrong_shape):
        message = f"parameter 'value': expected array with shape {correct_shape}, got {wrong_shape}"
        super(HubException, self).__init__(message=message)

class ModuleNotInstalledException(HubException):
    def __init__(self, module_name):
        message = f"Module '{module_name}' should be installed to convert the Dataset to the {module_name} format"
        super(HubException, self).__init__(message=message)

class WrongUsernameException(HubException):
    def __init__(self, username):
        message = f"The username {username} was not found. Make sure that the username provided in the url " \
                   "matches the one used during login."
        super(HubException, self).__init__(message=message)

class NotHubDatasetToOverwriteException(HubException):
    def __init__(self):
        message = "Unable to overwrite the dataset. "  \
                "The provided directory is not empty and doesn't contain information about any Hub Dataset "
        super(HubException, self).__init__(message=message)

class NotHubDatasetToAppendException(HubException):
    def __init__(self):
        message = "Unable to append to the dataset. "  \
                  "The provided directory is not empty and doesn't contain information about any Hub Dataset "
        super(HubException, self).__init__(message=message)


class NotZarrFolderException(Exception):
    pass


class StorageTensorNotFoundException(Exception):
    pass


class DynamicTensorNotFoundException(Exception):
    pass


import os
import hub
from outdated import check_outdated
import subprocess
import pkg_resources

from hub.exceptions import HubException
from hub.log import logger


def get_cli_version():
    return "1.0.0"


def verify_cli_version():
    try:
        version = pkg_resources.get_distribution(hub.__name__).version
        is_outdated, latest_version = check_outdated(hub.__name__, version)
        if is_outdated:
            print(
                "\033[93m"
                + "Hub is out of date. Please upgrade the package by running `pip3 install --upgrade snark`"
                + "\033[0m"
            )
    except Exception as e:
        logger.error(str(e))


def check_program_exists(command):
    try:
        subprocess.call([command])
    except OSError as e:
        if e.errno == os.errno.ENOENT:
            return False
        else:
            # Something else went wrong while trying to run `wget`
            return True
    return True


def get_proxy_command(proxy):
    ssh_proxy = ""
    if proxy and proxy != " " and proxy != "None" and proxy != "":
        if check_program_exists("ncat"):
            ssh_proxy = '-o "ProxyCommand=ncat --proxy-type socks5 --proxy {} %h %p"'.format(
                proxy
            )
        else:
            raise HubException(
                message="This pod is behind the firewall. You need one more thing. Please install nmap by running `sudo apt-get install nmap` on Ubuntu or `brew install nmap` for Mac"
            )
    return ssh_proxy



import hub.features.feature_connector as feature_connector

FeatureConnector = feature_connector.FeatureConnector


class Primitive(FeatureConnector):
    def __init__(self, dtype: str):
        self.dtype = dtype

    def _flatten(self):
        return "", (), self.dtype



from typing import Tuple
import numpy as np

from hub.features.features import Tensor
from hub.features.class_label import ClassLabel


class Segmentation(Tensor):
    """`FeatureConnector` for segmentation
    """
    def __init__(
        self,
        shape: Tuple[int, ...] = None,
        dtype: str = None,
        num_classes: int = None,
        names: Tuple[str] = None,
        names_file: str = None,
        max_shape: Tuple[int, ...] = None,
        chunks=True
    ):
        """Constructs a Segmentation FeatureConnector.
        Also constructs ClassLabel FeatureConnector for Segmentation classes.
        Args:
        shape: tuple of ints or None: (height, width, 1)
        dtype: dtype of segmentation array: `uint16` or `uint8`
        num_classes: `int`, number of classes. All labels must be < num_classes.
        names: `list<str>`, string names for the integer classes. The
                order in which the names are provided is kept.
        names_file: `str`, path to a file with names for the integer
                    classes, one per line.
        """
        super(Segmentation, self).__init__(shape, dtype, max_shape=max_shape, chunks=chunks)
        self.class_labels = ClassLabel(num_classes=num_classes, names=names, names_file=names_file, chunks=chunks)

    def get_segmentation_classes(self):
        class_indices = np.unique(self)
        return [self.class_labels.int2str(value) for value in class_indices]

    def get_attr_dict(self):
        """Return class attributes."""
        return self.__dict__



from typing import Tuple
import posixpath
import collections.abc as abc
import json
import sys
import os

import fsspec

from hub.features.features import (
    Primitive,
    Tensor,
    FeatureDict,
    FeatureConnector,
    featurify,
    FlatTensor,
)

from hub.api.tensorview import TensorView
from hub.api.datasetview import DatasetView
from hub.api.dataset_utils import slice_extract_info, slice_split
from hub.utils import compute_lcm

import hub.features.serialize
import hub.features.deserialize

from hub.store.dynamic_tensor import DynamicTensor
from hub.store.store import get_fs_and_path, get_storage_map
from hub.exceptions import (
    HubDatasetNotFoundException,
    NotHubDatasetToOverwriteException,
    NotHubDatasetToAppendException,
    ShapeArgumentNotFoundException,
    SchemaArgumentNotFoundException,
    ModuleNotInstalledException,
    WrongUsernameException
)
from hub.store.metastore import MetaStorage
from hub.client.hub_control import HubControlClient
try:
    import torch
except ImportError:
    pass
try:
    import tensorflow as tf
except ImportError:
    pass


def get_file_count(fs: fsspec.AbstractFileSystem, path):
    return len(fs.listdir(path, detail=False))


class Dataset:
    def __init__(
        self,
        url: str,
        mode: str = "a",
        safe_mode: bool = False,
        shape=None,
        schema=None,
        token=None,
        fs=None,
        fs_map=None,
        cache: int = 2 ** 26,
        lock_cache=True,
    ):
        """Open a new or existing dataset for read/write

        Parameters
        ----------
        url: str
            The url where dataset is located/should be created
        mode: str, optional (default to "w")
            Python way to tell whether dataset is for read or write (ex. "r", "w", "a")
        safe_mode: bool, optional
            if dataset exists it cannot be rewritten in safe mode, otherwise it lets to write the first time
        shape: tuple, optional
            Tuple with (num_samples,) format, where num_samples is number of samples
        schema: optional
            Describes the data of a single sample. Hub features are used for that
            Required for 'a' and 'w' modes
        token: str or dict, optional
            If url is refering to a place where authorization is required,
            token is the parameter to pass the credentials, it can be filepath or dict
        fs: optional
        fs_map: optional
        cache: int, optional
            Size of the cache. Default is 2GB (2**20)
        lock_cache: bool, optional
            Lock the cache for avoiding multiprocessing errors
        """

        shape = shape or (None,)
        if isinstance(shape, int): 
            shape = [shape]
        if shape is not None:
            assert len(tuple(shape)) == 1
        assert mode is not None

        self.url = url
        self.token = token
        self.mode = mode

        self._fs, self._path = (
            (fs, url) if fs else get_fs_and_path(self.url, token=token)
        )
        self.cache = cache
        self.lock_cache = lock_cache

        needcreate = self._check_and_prepare_dir()
        fs_map = fs_map or get_storage_map(self._fs, self._path, cache, lock=lock_cache)
        self._fs_map = fs_map

        if safe_mode and not needcreate:
            mode = "r"

        if not needcreate:
            self.meta = json.loads(fs_map["meta.json"].decode("utf-8"))
            self.shape = tuple(self.meta["shape"])
            self.schema = hub.features.deserialize.deserialize(self.meta["schema"])
            self._flat_tensors: Tuple[FlatTensor] = tuple(self.schema._flatten())
            self._tensors = dict(self._open_storage_tensors())
        else:
            try:
                if shape[0] is None:
                    raise ShapeArgumentNotFoundException()
                if schema is None:
                    raise SchemaArgumentNotFoundException()
                self.schema: FeatureConnector = featurify(schema)
                self.shape = tuple(shape)
                self.meta = {
                    "shape": shape,
                    "schema": hub.features.serialize.serialize(self.schema),
                    "version": 1,
                }
                fs_map["meta.json"] = bytes(json.dumps(self.meta), "utf-8")
                self._flat_tensors: Tuple[FlatTensor] = tuple(self.schema._flatten())
                self._tensors = dict(self._generate_storage_tensors())
            except Exception:
                self._fs.rm(self._path, recursive=True)
                raise
        self.username = None
        self.dataset_name = None
        if self._path.startswith("s3://snark-hub-dev/") or self._path.startswith("s3://snark-hub/"):
            subpath = self._path[5:]
            spl = subpath.split('/')
            if len(spl) < 4:
                raise ValueError("Invalid Path for dataset")
            self.username = spl[-2]
            self.dataset_name = spl[-1]
            if "w" in mode:
                HubControlClient().create_dataset_entry(self.username, self.dataset_name, meta)
                HubControlClient().update_dataset_state(self.username, self.dataset_name, "CREATED")

    def _check_and_prepare_dir(self):
        """
        Checks if input data is ok
        Creates or overwrites dataset folder
        Returns True dataset needs to be created opposed to read
        """
        fs, path, mode = self._fs, self._path, self.mode
        if path.startswith('s3://'):
            with open(os.path.expanduser('~/.activeloop/store'), 'rb') as f:
                stored_username = json.load(f)['_id']
            current_username = path.split('/')[-2]
            if stored_username != current_username:
                raise WrongUsernameException(current_username)
        exist_meta = fs.exists(posixpath.join(path, "meta.json"))
        if exist_meta:
            if "w" in mode:
                fs.rm(path, recursive=True)
                fs.makedirs(path)
                return True
            return False
        else:
            if "r" in mode:
                raise HubDatasetNotFoundException(path)
            exist_dir = fs.exists(path)
            if not exist_dir:
                fs.makedirs(path)
            elif get_file_count(fs, path) > 0:
                if "w" in mode:
                    raise NotHubDatasetToOverwriteException()
                else:
                    raise NotHubDatasetToAppendException()
            return True

    def _generate_storage_tensors(self):
        for t in self._flat_tensors:
            t: FlatTensor = t
            path = posixpath.join(self._path, t.path[1:])
            self._fs.makedirs(posixpath.join(path, "--dynamic--"))
            yield t.path, DynamicTensor(
                fs_map=MetaStorage(
                    t.path,
                    get_storage_map(self._fs, path, self.cache, self.lock_cache),
                    self._fs_map,
                ),
                mode=self.mode,
                shape=self.shape + t.shape,
                max_shape=self.shape + t.max_shape,
                dtype=t.dtype,
                chunks=t.chunks,
            )

    def _open_storage_tensors(self):
        for t in self._flat_tensors:
            t: FlatTensor = t
            path = posixpath.join(self._path, t.path[1:])
            yield t.path, DynamicTensor(
                fs_map=MetaStorage(
                    t.path,
                    get_storage_map(self._fs, path, self.cache, self.lock_cache),
                    self._fs_map,
                ),
                mode=self.mode,
                shape=self.shape + t.shape,
            )

    def __getitem__(self, slice_):
        """Gets a slice or slices from dataset
        Examples
        --------
        return ds["image", 5, 0:1920, 0:1080, 0:3].compute() # returns numpy array

        images = ds["image"]
        return images[5].compute() # returns numpy array

        images = ds["image"]
        image = images[5]
        return image[0:1920, 0:1080, 0:3].compute()
        """
        if not isinstance(slice_, abc.Iterable) or isinstance(slice_, str):
            slice_ = [slice_]
        slice_ = list(slice_)
        subpath, slice_list = slice_split(slice_)
        if not subpath:
            if len(slice_list) > 1:
                raise ValueError(
                    "Can't slice a dataset with multiple slices without subpath"
                )
            num, ofs = slice_extract_info(slice_list[0], self.shape[0])
            return DatasetView(dataset=self, num_samples=num, offset=ofs)
        elif not slice_list:
            if subpath in self._tensors.keys():
                return TensorView(
                    dataset=self, subpath=subpath, slice_=slice(0, self.shape[0])
                )
            return self._get_dictionary(subpath)
        else:
            num, ofs = slice_extract_info(slice_list[0], self.shape[0])
            if subpath in self._tensors.keys():
                return TensorView(dataset=self, subpath=subpath, slice_=slice_list)
            if len(slice_list) > 1:
                raise ValueError("You can't slice a dictionary of Tensors")
            return self._get_dictionary(subpath, slice_list[0])

    def __setitem__(self, slice_, value):
        """ "Sets a slice or slices with a value
        Examples
        --------
        ds["image", 5, 0:1920, 0:1080, 0:3] = np.zeros((1920, 1080, 3), "uint8")

        images = ds["image"]
        image = images[5]
        image[0:1920, 0:1080, 0:3] = np.zeros((1920, 1080, 3), "uint8")
        """
        if not isinstance(slice_, abc.Iterable) or isinstance(slice_, str):
            slice_ = [slice_]
        slice_ = list(slice_)
        subpath, slice_list = slice_split(slice_)
        if not subpath:
            raise ValueError("Can't assign to dataset sliced without subpath")
        elif not slice_list:
            self._tensors[subpath][:] = value  # Add path check
        else:
            self._tensors[subpath][slice_list] = value

    def delete(self):
        fs, path = self._fs, self._path
        exist_meta = fs.exists(posixpath.join(path, "meta.json"))
        if exist_meta:
            fs.rm(path, recursive=True)
            if self.username is not None:
                HubControlClient().delete_dataset_entry(self.username, self.dataset_name)
            return True
        return False

    def to_pytorch(self, Transform=None):
        if "torch" not in sys.modules:
            raise ModuleNotInstalledException('torch')
        return TorchDataset(self, Transform)

    def to_tensorflow(self):
        if "tensorflow" not in sys.modules:
            raise ModuleNotInstalledException('tensorflow')
        def tf_gen():
            for index in range(self.shape[0]):
                d = {}
                for key in self._tensors.keys():
                    split_key = key.split("/")
                    cur = d
                    for i in range(1, len(split_key) - 1):
                        if split_key[i] in cur.keys():
                            cur = cur[split_key[i]]
                        else:
                            cur[split_key[i]] = {}
                            cur = cur[split_key[i]]
                    cur[split_key[-1]] = self._tensors[key][index]
                yield (d)

        def dict_to_tf(my_dtype):
            d = {}
            for k, v in my_dtype.dict_.items():
                d[k] = dtype_to_tf(v)
            return d

        def tensor_to_tf(my_dtype):
            return dtype_to_tf(my_dtype.dtype)

        def dtype_to_tf(my_dtype):
            if isinstance(my_dtype, FeatureDict):
                return dict_to_tf(my_dtype)
            elif isinstance(my_dtype, Tensor):
                return tensor_to_tf(my_dtype)
            elif isinstance(my_dtype, Primitive):
                return str(my_dtype._dtype)

        output_types = dtype_to_tf(self.schema)
        return tf.data.Dataset.from_generator(
            tf_gen,
            output_types=output_types,
        )

    def _get_dictionary(self, subpath, slice_=None):
        """"Gets dictionary from dataset given incomplete subpath"""
        tensor_dict = {}
        subpath = subpath if subpath.endswith("/") else subpath + "/"
        for key in self._tensors.keys():
            if key.startswith(subpath):
                suffix_key = key[len(subpath) :]
                split_key = suffix_key.split("/")
                cur = tensor_dict
                for i in range(len(split_key) - 1):
                    if split_key[i] not in cur.keys():
                        cur[split_key[i]] = {}
                    cur = cur[split_key[i]]
                slice_ = slice_ if slice_ else slice(0, self.shape[0])
                cur[split_key[-1]] = TensorView(
                    dataset=self, subpath=key, slice_=slice_
                )
        if len(tensor_dict) == 0:
            raise KeyError(f"Key {subpath} was not found in dataset")
        return tensor_dict

    def __iter__(self):
        """ Returns Iterable over samples """
        for i in range(len(self)):
            yield self[i]

    def __len__(self):
        """ Number of samples in the dataset """
        return self.shape[0]

    def commit(self):
        """Save changes from cache to dataset final storage
        This invalidates this object
        """
        for t in self._tensors.values():
            t.commit()
        if self.username is not None:
            HubControlClient().update_dataset_state(self.username, self.dataset_name, "UPLOADED")

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, exc_traceback):
        self.commit()

    @property
    def chunksize(self):
        # FIXME assumes chunking is done on the first sample
        chunks = [t.chunksize[0] for t in self._tensors.values()]
        return compute_lcm(chunks)

    @property
    def keys(self):
        """
        Get Keys of the dataset
        """
        return self._tensors.keys()


class TorchDataset:
    def __init__(self, ds, transform=None):
        self._ds = None
        self._url = ds.url
        self._token = ds.token
        self._transform = transform

    def _do_transform(self, data):
        return self._transform(data) if self._transform else data

    def _init_ds(self):
        """
        For each process, dataset should be independently loaded
        """
        if self._ds is None:
            self._ds = Dataset(self._url, token=self._token, lock_cache=False)

    def __len__(self):
        self._init_ds()
        return self._ds.shape[0]

    def __getitem__(self, index):
        self._init_ds()
        d = {}
        for key in self._ds._tensors.keys():
            split_key = key.split("/")
            cur = d
            for i in range(1, len(split_key) - 1):
                if split_key[i] in cur.keys():
                    cur = cur[split_key[i]]
                else:
                    cur[split_key[i]] = {}
                    cur = cur[split_key[i]]

            cur[split_key[-1]] = torch.tensor(self._ds._tensors[key][index])
        return d

    def __iter__(self):
        self._init_ds()
        for index in range(self.shape[0]):
            d = {}
            for key in self._ds._tensors.keys():
                split_key = key.split("/")
                cur = d
                for i in range(1, len(split_key) - 1):
                    if split_key[i] in cur.keys():
                        cur = cur[split_key[i]]
                    else:
                        cur[split_key[i]] = {}
                        cur = cur[split_key[i]]
                cur[split_key[-1]] = torch.tensor(self._tensors[key][index])
            yield (d)


import posixpath

import numpy as np
import fsspec

from hub.store.dynamic_tensor import DynamicTensor
from hub.store.store import StorageMapWrapperWithCommit


def create_store(path: str):
    fs: fsspec.AbstractFileSystem = fsspec.filesystem("file")
    if fs.exists(path):
        fs.rm(path, recursive=True)
    fs.makedirs(posixpath.join(path, "--dynamic--"))
    mapper = fs.get_mapper(path)
    mapper["--dynamic--/hello.txt"] = bytes("Hello World", "utf-8")
    return StorageMapWrapperWithCommit(mapper)


def test_dynamic_tensor():
    t = DynamicTensor(
        create_store("./data/test/test_dynamic_tensor"),
        mode="w",
        shape=(5, 100, 100),
        max_shape=(5, 100, 100),
        dtype="int32",
    )
    t[0, 80:, 80:] = np.ones((20, 20), dtype="int32")
    assert t[0, -5, 90:].tolist() == [1] * 10


def test_dynamic_tensor_2():
    t = DynamicTensor(
        create_store("./data/test/test_dynamic_tensor_2"),
        mode="w",
        shape=(5, None, None),
        max_shape=(5, 100, 100),
        dtype="int32",
    )
    t[0] = np.ones((10, 10), dtype="int32")
    assert t[0, 5].tolist() == [1] * 10
    assert t[0, 5, :].tolist() == [1] * 10
    t[0, 6] = 2 * np.ones((20,), dtype="int32")
    assert t[0, 5, :].tolist() == [1] * 10 + [0] * 10


def test_dynamic_tensor_3():
    t = DynamicTensor(
        create_store("./data/test/test_dynamic_tensor_3"),
        mode="w",
        shape=(5, None, None, None),
        max_shape=(5, 100, 100, 100),
        dtype="int32",
    )
    t[0, 5] = np.ones((20, 30), dtype="int32")
    t[0, 6:8, 5:9] = 5 * np.ones((2, 4, 30), dtype="int32")
    assert t[0, 5, 7].tolist() == [1] * 30
    assert t[0, 7, 8].tolist() == [5] * 30


def test_dynamic_tensor_shapes():
    t = DynamicTensor(
        create_store("./data/test/test_dynamic_tensor_5"),
        mode="w",
        shape=(5, None, None),
        max_shape=(5, 100, 100),
        dtype="int32",
    )
    t[0] = np.ones((5, 10), dtype="int32")
    t[0, 6] = 2 * np.ones((20,), dtype="int32")
    assert t[0, -1].tolist() == [2] * 20


def test_dynamic_tensor_4():
    t = DynamicTensor(
        create_store("./data/test/test_dynamic_tensor_6"),
        mode="w",
        shape=(5, None, None, None),
        max_shape=(5, 100, 100, 10),
        dtype="int32",
    )
    t[0, 6:8] = np.ones((2, 20, 10), dtype="int32")
    assert (t[0, 6:8] == np.ones((2, 20, 10), dtype="int32")).all()


def test_chunk_iterator():
    t = DynamicTensor(
        create_store("./data/test/test_dynamic_tensor_7"),
        mode="w",
        shape=(50, 100, 100, 100),
        max_shape=(50, 100, 100, 100),
        dtype="int32",
    )

    assert t.chunksize == (5, 100, 100, 100)
    assert list(t.chunk_iterator())[0].shape == t.chunksize


if __name__ == "__main__":
    test_dynamic_tensor_2()
    # test_chunk_iterator()
    # test_dynamic_tensor_shapes()


from typing import Tuple
# import numpy as np

from hub.features.features import Tensor
# from hub.features.polygon import Polygon


class Mask(Tensor):
    """`FeatureConnector` for mask
    """
    def __init__(
        self,
        shape: Tuple[int, ...] = None,
        dtype=None,
        max_shape: Tuple[int, ...] = None,
        chunks=True
    ):
        """Constructs a Mask FeatureConnector.

        shape: tuple of ints or None: (height, width, 1)
        dtype: dtype of mask array. Default: `uint8`

        Example:
        ```python
        mask_tensor = Mask(shape=(300, 300, 1))
        ```
        """
        if not dtype:
            dtype = 'uint8'
        super(Mask, self).__init__(shape, dtype, max_shape=max_shape, chunks=chunks)

    def get_attr_dict(self):
        """Return class attributes."""
        return self.__dict__



from hub.features.features import Tensor


class Sequence(Tensor):
    """`Sequence` correspond to sequence of `features.FeatureConnector`.
    At generation time, a list for each of the sequence element is given. The output
    of `Dataset` will batch all the elements of the sequence together.
    If the length of the sequence is static and known in advance, it should be
    specified in the constructor using the `length` param.

    Example:
    ```
    sequence = Sequence(Image(), length=NB_FRAME)
    ```
    """

    def __init__(self, feature: Tensor, length: int = None, chunks=True):
        """Construct a sequence of Tensors.
        Args:
        feature: the features to wrap
        length: `int`, length of the sequence if static and known in advance
        """
        super(Sequence, self).__init__(shape=(length,), dtype=feature, chunks=chunks)

    def get_attr_dict(self):
        """Return class attributes"""
        return self.__dict__
